//////////////////////////////////////////////////////////////////////////////
// DEC.h
// Types for forms and operators.
// These are all just Vectors and SparseMatrices, but with these derived
// classes we can enforce semantics on the dimensions. I.e. Enforce that d of
// a k-form is a (k+1)-form, etc.
//////////////////////////////////////////////////////////////////////////////

#ifndef _DEC_H_
#define _DEC_H_


//////////////////////////////////////////////////////////////////////////////
// Includes

#include "simplicialComplex/simplicialComplex.h"
#include "linearAlgebra/linearAlgebra.h"


//////////////////////////////////////////////////////////////////////////////

enum FormType { PRIMAL, DUAL };


//////////////////////////////////////////////////////////////////////////////
// class Form
// T specifies whether it's a primal form or a dual form, and N is the dimension

template<FormType T, int N>
class Form : public Vector
{
public:
    Form() : Vector(0) { }

    Form( const SimplicialComplex<2> &mesh )
        : Vector( ( (T == PRIMAL && N == 0) || (T == DUAL && N == 2) ) ? (int)mesh.NumVerts() :
                ( ( (T == PRIMAL && N == 1) || (T == DUAL && N == 1) ) ? (int)mesh.NumEdges() :
                ( ( (T == PRIMAL && N == 2) || (T == DUAL && N == 0) ) ? (int)mesh.NumFaces() : 0 ) ) )
    { }

    Form( const SimplicialComplex<3> &mesh )
        : Vector( ( (T == PRIMAL && N == 0) || (T == DUAL && N == 3) ) ? (int)mesh.NumVerts() :
                ( ( (T == PRIMAL && N == 1) || (T == DUAL && N == 2) ) ? (int)mesh.NumEdges() :
                ( ( (T == PRIMAL && N == 2) || (T == DUAL && N == 1) ) ? (int)mesh.NumFaces() :
                ( ( (T == PRIMAL && N == 3) || (T == DUAL && N == 0) ) ? (int)mesh.NumTets()  : 0 ) ) ) )
    { }

    void Init( const SimplicialComplex<2> &mesh )
    {
        Vector::Assign( ( (T == PRIMAL && N == 0) || (T == DUAL && N == 2) ) ? (int)mesh.NumVerts() :
                      ( ( (T == PRIMAL && N == 1) || (T == DUAL && N == 1) ) ? (int)mesh.NumEdges() :
                      ( ( (T == PRIMAL && N == 2) || (T == DUAL && N == 0) ) ? (int)mesh.NumFaces() : 0 ) ), 0.0f );
    }

    void Init( const SimplicialComplex<3> &mesh )
    {
        Vector::Assign( ( (T == PRIMAL && N == 0) || (T == DUAL && N == 3) ) ? (int)mesh.NumVerts() :
                      ( ( (T == PRIMAL && N == 1) || (T == DUAL && N == 2) ) ? (int)mesh.NumEdges() :
                      ( ( (T == PRIMAL && N == 2) || (T == DUAL && N == 1) ) ? (int)mesh.NumFaces() :
                      ( ( (T == PRIMAL && N == 3) || (T == DUAL && N == 0) ) ? (int)mesh.NumTets()  : 0 ) ) ), 0.0f );
    }

private:
    // disable resizing.
    void Assign( int, double );
};


//////////////////////////////////////////////////////////////////////////////
// class DecOp
// T1/N1 specify the operand, and T2/N2 specify the result

template<FormType T1, int N1, FormType T2, int N2>
class DecOp : public SparseMatrix
{
public:
    void BeginConstruction( const SimplicialComplex<2> &mesh, bool bSymmetric = false )
    {
        int numCols( ( (T1 == PRIMAL && N1 == 0) || (T1 == DUAL && N1 == 2) ) ? (int)mesh.NumVerts() :
                   ( ( (T1 == PRIMAL && N1 == 1) || (T1 == DUAL && N1 == 1) ) ? (int)mesh.NumEdges() :
                   ( ( (T1 == PRIMAL && N1 == 2) || (T1 == DUAL && N1 == 0) ) ? (int)mesh.NumFaces() : 0 ) ) );

        int numRows( ( (T2 == PRIMAL && N2 == 0) || (T2 == DUAL && N2 == 2) ) ? (int)mesh.NumVerts() :
                   ( ( (T2 == PRIMAL && N2 == 1) || (T2 == DUAL && N2 == 1) ) ? (int)mesh.NumEdges() :
                   ( ( (T2 == PRIMAL && N2 == 2) || (T2 == DUAL && N2 == 0) ) ? (int)mesh.NumFaces() : 0 ) ) );

        SparseMatrix::BeginConstruction( numRows, numCols, bSymmetric );
    }

    void BeginConstruction( const SimplicialComplex<3> &mesh, bool bSymmetric = false )
    {
        int numCols( ( (T1 == PRIMAL && N1 == 0) || (T1 == DUAL && N1 == 3) ) ? (int)mesh.NumVerts() :
                   ( ( (T1 == PRIMAL && N1 == 1) || (T1 == DUAL && N1 == 2) ) ? (int)mesh.NumEdges() :
                   ( ( (T1 == PRIMAL && N1 == 2) || (T1 == DUAL && N1 == 1) ) ? (int)mesh.NumFaces() :
                   ( ( (T1 == PRIMAL && N1 == 3) || (T1 == DUAL && N1 == 0) ) ? (int)mesh.NumTets()  : 0 ) ) ) );

        int numRows( ( (T2 == PRIMAL && N2 == 0) || (T2 == DUAL && N2 == 3) ) ? (int)mesh.NumVerts() :
                   ( ( (T2 == PRIMAL && N2 == 1) || (T2 == DUAL && N2 == 2) ) ? (int)mesh.NumEdges() :
                   ( ( (T2 == PRIMAL && N2 == 2) || (T2 == DUAL && N2 == 1) ) ? (int)mesh.NumFaces() :
                   ( ( (T2 == PRIMAL && N2 == 3) || (T2 == DUAL && N2 == 0) ) ? (int)mesh.NumTets()  : 0 ) ) ) );

        SparseMatrix::BeginConstruction( numRows, numCols, bSymmetric );
    }

    void Apply( const Form<T1,N1> &vec, Form<T2,N2> &product ) const
    {
        SparseMatrix::Multiply( vec, product );
    }

    bool ApplyInverse( const Form<T2,N2> &b, Form<T1,N1> &x ) const
    {
        return SparseMatrix::Solve( b, x );
    }

private:
    // disable
    void Multiply( const Vector &, Vector & );
    bool Solve( const Vector &, Vector &, int, real );
};


#endif  // _DEC_H_
