
//////////////////////////////////////////////////////////////////////////////
// Includes

#include "fluids/octree.h"
#include "simplicialComplex/simplicialComplex.h"
#include "simplicialComplex/iterator.h"
#include "simplicialComplex/vertex.h"
#include "simplicialComplex/tet.h"
#include "util/geometry.h"
#include "util/bbox.h"
#include "util/assert.h"

#include <vector>
#include <limits>

using namespace std;


const size_t MAX_LEAF_ELEMENTS = 10;
const size_t MAX_OCTREE_LEVELS = 5;


//////////////////////////////////////////////////////////////////////////////
// class OctNode

class OctNode
{
    friend class Octree;

private:
    OctNode( const BBox& bbox, const std::vector<BBox>& tetBounds, const std::vector<size_t>& activeTets );
    ~OctNode();

private:
    void Build( const BBox& bbox, const std::vector<BBox>& tetBounds, const std::vector<size_t>& activeTets );

    OctNode * Locate( const vec3& point );

    vector<size_t> m_enclosedTets;
    BBox           m_bbox;
    bool           m_bLeaf;
    OctNode       *m_children[8];
};

//////////////////////////////////////////////////////////////////////////////
// Default Constructor

OctNode::OctNode( const BBox& bbox, const std::vector<BBox>& tetBounds, const std::vector<size_t>& activeTets )
{
    m_bLeaf = false;
    m_bbox.SetBounds( vec3(numeric_limits<real>::max()), vec3(numeric_limits<real>::min()) );
    for( int i = 0; i < 8; ++i )
        m_children[i] = 0;

    Build( bbox, tetBounds, activeTets );
}

//////////////////////////////////////////////////////////////////////////////
// Destructor

OctNode::~OctNode()
{
    for( int i = 0; i < 8; ++i )
        delete m_children[i];
}

//////////////////////////////////////////////////////////////////////////////
// Build - recursively fill the tree

void OctNode::Build( const BBox& bbox, const vector<BBox>& tetBounds, const vector<size_t>& activeTets )
{
    static size_t level = 0;

    m_bbox = bbox;

    // record which tets are within the bounds of this node
    for( size_t tet = 0; tet < activeTets.size(); ++tet )
    {
        if( m_bbox.Intersect( tetBounds[ activeTets[tet] ] ) ) 
            m_enclosedTets.push_back( activeTets[tet] );
    }

    // recurse if we have too many elements, but only if we're not too deep
    if( m_enclosedTets.size() > MAX_LEAF_ELEMENTS && level < MAX_OCTREE_LEVELS )
    {
        BBox box;
        vec3 xSplit( m_bbox.GetMin()[0], m_bbox.GetCenter()[0], m_bbox.GetMax()[0] );
        vec3 ySplit( m_bbox.GetMin()[1], m_bbox.GetCenter()[1], m_bbox.GetMax()[1] );
        vec3 zSplit( m_bbox.GetMin()[2], m_bbox.GetCenter()[2], m_bbox.GetMax()[2] );
        for( int i = 0; i < 2; ++i )
        {
            for( int j = 0; j < 2; ++j )
            {
                for( int k = 0; k < 2; ++k )
                {
                    assert( m_children[4*i + 2*j + k] == 0 );

                    box.SetBounds( xSplit[i], ySplit[j], zSplit[k], xSplit[i+1], ySplit[j+1], zSplit[k+1] );

                    ++level;
                    m_children[4*i + 2*j + k] = new OctNode( box, tetBounds, m_enclosedTets );
                    --level;
                }
            }
        }
    }
    else
    {
        m_bLeaf = true;
    }
}

//////////////////////////////////////////////////////////////////////////////
// Locate - find which node this point is in.

OctNode * OctNode::Locate( const vec3& point )
{
    if( !m_bLeaf )
    {
        const vec3 boxCenter = m_bbox.GetCenter();

        int whichChild = 0;
        if( point[0] > boxCenter[0] ) whichChild += 4;
        if( point[1] > boxCenter[1] ) whichChild += 2;
        if( point[2] > boxCenter[2] ) whichChild += 1;

        return m_children[ whichChild ]->Locate( point );
    }

    // this is a leaf
    return this;
}


//////////////////////////////////////////////////////////////////////////////
// Octree
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
// Default Constructor

Octree::Octree( const SimplicialComplex<3> *sc ) : sc(sc), m_root(0)
{
    Build();
}

//////////////////////////////////////////////////////////////////////////////
// Destructor

Octree::~Octree()
{
    delete m_root;
}

//////////////////////////////////////////////////////////////////////////////
// Build

void Octree::Build()
{
    // First find the bounding box of the entire domain
    vec3 bboxMin( numeric_limits<real>::max() );
    vec3 bboxMax( numeric_limits<real>::min() );
    for( SimplexIterator<Vertex> vertex = sc->Vertices(); vertex; ++vertex )
    {
        vec3 pos = vertex->Position();
        bboxMin[0] = min( bboxMin[0], pos[0] );
        bboxMin[1] = min( bboxMin[1], pos[1] );
        bboxMin[2] = min( bboxMin[2], pos[2] );
        bboxMax[0] = max( bboxMax[0], pos[0] );
        bboxMax[1] = max( bboxMax[1], pos[1] );
        bboxMax[2] = max( bboxMax[2], pos[2] );
    }

    BBox rootNodeBound( bboxMin, bboxMax );

    // This vector specifies which tets are relevant to the current node.
    // For example, when building a child we only need to consider the tets
    // that were present in the parent.
    vector<size_t> activeTets( sc->NumTets() );

    // Make a bounding box for each tet. The rest of the tree construction
    // process will operate on these bounds instead of the tets.
    vector<BBox> tetBounds( sc->NumTets() );
    vec3 v0, v1, v2, v3;
    for( SimplexIterator<Tet> tet = sc->Tets(); tet; ++tet )
    {
        tet->Vertices( v0, v1, v2, v3 );

        bboxMin[0] = min( v0[0], min( v1[0], min( v2[0], v3[0] ) ) );
        bboxMin[1] = min( v0[1], min( v1[1], min( v2[1], v3[1] ) ) );
        bboxMin[2] = min( v0[2], min( v1[2], min( v2[2], v3[2] ) ) );
        bboxMax[0] = max( v0[0], max( v1[0], max( v2[0], v3[0] ) ) );
        bboxMax[1] = max( v0[1], max( v1[1], max( v2[1], v3[1] ) ) );
        bboxMax[2] = max( v0[2], max( v1[2], max( v2[2], v3[2] ) ) );

        tetBounds[ tet->Index() ].SetBounds( bboxMin, bboxMax );
        activeTets[ tet->Index() ] = tet->Index();
    }

    // Build the tree
    smart_assert( m_root == 0 ).fatal();
    m_root = new OctNode( rootNodeBound, tetBounds, activeTets );
}

//////////////////////////////////////////////////////////////////////////////
// Locate - find which tet this point is in

Tet Octree::Locate( const vec3& point )
{
    // First find which node the point is in
    OctNode *node = m_root->Locate( point );
    if( !node )
        return Tet();
    smart_assert( node->m_bLeaf ).fatal();

    // Then test against all tets that intersect that node
    Tet tet;
    vec3 v0, v1, v2, v3;
    for( size_t i = 0; i < node->m_enclosedTets.size(); ++i )
    {
        tet = sc->GetTet( node->m_enclosedTets[i] );
        tet.Vertices( v0, v1, v2, v3 );
        if( IsPointInTet(point,v0,v1,v2,v3) )
            return tet;
    }

    return Tet();
}
