//////////////////////////////////////////////////////////////////////////////
// LinAlgPetscImpl.cpp
// Petsc implementations of the Vector and Sparse Matrix classes.
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Includes

#ifdef COMPILE_WITH_PETSC

#include "linearAlgebra/linearAlgebra.h"
#include "linearAlgebra/linAlgImpl.h"
#include "util/assert.h"

using namespace std;


//////////////////////////////////////////////////////////////////////////////

#undef CHKERRQ
#define CHKERRQ(n) if (n) {LOG(err) << "Petsc Error " << n << ", on line " << __LINE__;}


//////////////////////////////////////////////////////////////////////////////
// Initializes Petsc at the start of the program execution

struct _PetscInitializer
{
    _PetscInitializer() {
        PetscTruth bInitialized = PETSC_FALSE;
        PetscErrorCode ierr = PetscInitialized(&bInitialized);CHKERRQ(ierr);
        ierr = bInitialized ? 0 : PetscInitializeNoArguments();CHKERRQ(ierr);
    }
} PetscInitializer;


//////////////////////////////////////////////////////////////////////////////
// PetscVectorImpl

PetscVectorImpl::PetscVectorImpl( size_t size ) : data(0)
{
    PetscErrorCode ierr;
    ierr = VecCreateSeq( MPI_COMM_SELF, (PetscInt)size, &m_PetscVec );CHKERRQ(ierr);
    ierr = VecGetArray( m_PetscVec, &data );CHKERRQ(ierr);
}
PetscVectorImpl::~PetscVectorImpl()
{
    PetscErrorCode ierr = VecDestroy( m_PetscVec );CHKERRQ(ierr);
}
double & PetscVectorImpl::operator [] ( size_t i )
{
    return data[i];
}
const double & PetscVectorImpl::operator [] ( size_t i ) const
{
    return data[i];
}
void PetscVectorImpl::Assign( size_t size, double value )
{
    PetscErrorCode ierr;
    ierr = VecDestroy( m_PetscVec );CHKERRQ(ierr);
    ierr = VecCreateSeq( MPI_COMM_SELF, (PetscInt)size, &m_PetscVec );CHKERRQ(ierr);
    ierr = VecSet( &value, m_PetscVec );CHKERRQ(ierr);
    ierr = VecGetArray( m_PetscVec, &data );CHKERRQ(ierr);
}
void PetscVectorImpl::Set( double value )
{
    PetscErrorCode ierr = VecSet( &value, m_PetscVec );CHKERRQ(ierr);
}
void PetscVectorImpl::Copy( const VectorImpl *v )
{
    PetscErrorCode ierr = VecCopy( static_cast<const PetscVectorImpl*>(v)->m_PetscVec, m_PetscVec );CHKERRQ(ierr);
}
void PetscVectorImpl::Swap( VectorImpl *v )
{
    PetscErrorCode ierr = VecSwap( static_cast<const PetscVectorImpl*>(v)->m_PetscVec, m_PetscVec );CHKERRQ(ierr);
}
void PetscVectorImpl::Add( const VectorImpl *v )
{
    static const double one = 1.0;
    PetscErrorCode ierr = VecAXPY( &one, static_cast<const PetscVectorImpl*>(v)->m_PetscVec, m_PetscVec );CHKERRQ(ierr);
}
void PetscVectorImpl::Subtract( const VectorImpl *v )
{
    static const double minusOne = -1.0;
    PetscErrorCode ierr = VecAXPY( &minusOne, static_cast<const PetscVectorImpl*>(v)->m_PetscVec, m_PetscVec );CHKERRQ(ierr);
}
void PetscVectorImpl::Multiply( const VectorImpl *v )
{
    PetscErrorCode ierr = VecPointwiseMult( m_PetscVec, static_cast<const PetscVectorImpl*>(v)->m_PetscVec, m_PetscVec );CHKERRQ(ierr);
}
void PetscVectorImpl::Multiply( double f )
{
    PetscErrorCode ierr = VecScale( &f, m_PetscVec );CHKERRQ(ierr);
}
void PetscVectorImpl::Divide( const VectorImpl *v )
{
    PetscErrorCode ierr = VecPointwiseDivide( m_PetscVec, static_cast<const PetscVectorImpl*>(v)->m_PetscVec, m_PetscVec );CHKERRQ(ierr);
}
void PetscVectorImpl::Negate( void )
{
    static const double minusOne = -1.0;
    PetscErrorCode ierr = VecScale( &minusOne, m_PetscVec );CHKERRQ(ierr);
}
void PetscVectorImpl::MAD( double f, const VectorImpl *v )
{
    PetscErrorCode ierr = VecAYPX( &f, static_cast<const PetscVectorImpl*>(v)->m_PetscVec, m_PetscVec );CHKERRQ(ierr);
}
void PetscVectorImpl::MAD( const VectorImpl *v, double f )
{
    PetscErrorCode ierr = VecAXPY( &f, static_cast<const PetscVectorImpl*>(v)->m_PetscVec, m_PetscVec );CHKERRQ(ierr);
}
real PetscVectorImpl::DotProduct( const VectorImpl *v ) const
{
    double result;
    PetscErrorCode ierr = VecDot( m_PetscVec, static_cast<const PetscVectorImpl*>(v)->m_PetscVec, &result );CHKERRQ(ierr);
    return (real)result;
}


//////////////////////////////////////////////////////////////////////////////
// Default Constructor

PetscMatImpl::PetscMatImpl()
    : m_numRows(0),
      m_numColumns(0),
      m_bSymmetric(false),
      m_PetscMat(0),
      m_PetscKSP(0)
{
    // empty
}

//////////////////////////////////////////////////////////////////////////////
// Constructor - takes matrix size as arguments

PetscMatImpl::PetscMatImpl( int numRows, int numColumns )
    : m_numRows(numRows),
      m_numColumns(numColumns),
      m_bSymmetric(false),
      m_rows(numRows),
      m_PetscMat(0),
      m_PetscKSP(0)
{
    // empty
}

//////////////////////////////////////////////////////////////////////////////
// Destructor

PetscMatImpl::~PetscMatImpl()
{
    Clear();
}

//////////////////////////////////////////////////////////////////////////////
// Clear

void PetscMatImpl::Clear()
{
    m_numRows = 0;
    m_numColumns = 0;
    m_bSymmetric = false;

    m_rows.clear();

    if( m_PetscKSP )
    {
        PetscErrorCode ierr = KSPDestroy(m_PetscKSP);CHKERRQ(ierr);
        m_PetscKSP = 0;
    }
    if( m_PetscMat )
    {
        PetscErrorCode ierr = MatDestroy(m_PetscMat);CHKERRQ(ierr);
        m_PetscMat = 0;
    }
}

//////////////////////////////////////////////////////////////////////////////
// BeginConstruction - Sets the size and prepares for the construction phase

void PetscMatImpl::BeginConstruction( int numRows, int numColumns, bool bSymmetric )
{
    Clear();

    if( numRows > 0 && numColumns > 0 )
    {
        m_numRows = numRows;
        m_numColumns = numColumns;
        m_rows.resize( numRows );
    }

    m_bSymmetric = bSymmetric;
}

//////////////////////////////////////////////////////////////////////////////
// EndConstruction - Wraps up the construction phase. Entries cannot be
//                   modified after this function has been called

void PetscMatImpl::EndConstruction( PRECOND_TYPE_ENUM, real )
{
    // Convert to Petsc representation
    int *nnz = new int[m_numRows];
    for( int i = 0; i < m_numRows; ++i )
        nnz[i] = (int)m_rows[i].size();

    PetscErrorCode ierr = MatCreateSeqAIJ( MPI_COMM_SELF, m_numRows, m_numColumns, 0, nnz, &m_PetscMat );CHKERRQ(ierr);

    int j = 0;
    for( int i = 0; i < m_numRows; ++i )
    {
        for( map<int,double>::const_iterator rowIter = m_rows[i].begin();
            rowIter != m_rows[i].end(); ++rowIter, ++j )
        {
            ierr = MatSetValue( m_PetscMat, i, rowIter->first, rowIter->second, INSERT_VALUES );CHKERRQ(ierr);
        }
    }

    ierr = MatAssemblyBegin( m_PetscMat, MAT_FINAL_ASSEMBLY );CHKERRQ(ierr);
    ierr = MatAssemblyEnd( m_PetscMat, MAT_FINAL_ASSEMBLY );CHKERRQ(ierr);

    // Free up the memory from the other representation
    m_rows.clear();

    // Initialize the solver
    InitKSP();
}

//////////////////////////////////////////////////////////////////////////////
// InitKSP - Initialize the preconditioner, etc

void PetscMatImpl::InitKSP()
{
    if( !m_bSymmetric )
        return;

    PetscErrorCode ierr;
    ierr = KSPCreate( MPI_COMM_SELF, &m_PetscKSP );CHKERRQ(ierr);
    ierr = KSPSetOperators( m_PetscKSP, m_PetscMat, m_PetscMat, SAME_PRECONDITIONER );CHKERRQ(ierr);
    ierr = KSPSetType( m_PetscKSP, KSPGMRES );CHKERRQ(ierr);
    ierr = KSPSetInitialGuessNonzero( m_PetscKSP, PETSC_TRUE );CHKERRQ(ierr);
    ierr = KSPSetTolerances( m_PetscKSP, 0.001, 1e-6, 100000, 10000 );CHKERRQ(ierr);

    PC pc;
    ierr = KSPGetPC( m_PetscKSP, &pc );CHKERRQ(ierr);
    ierr = PCSetType( pc, PCEISENSTAT );CHKERRQ(ierr);
    //ierr = PCEisenstatSetOmega( pc, 0.5 );CHKERRQ(ierr);
    //ierr = KSPSetComputeEigenvalues( m_PetscKSP, PETSC_TRUE );
    ierr = KSPSetUp( m_PetscKSP );CHKERRQ(ierr);
}

//////////////////////////////////////////////////////////////////////////////
// Save - Save to file in Petsc's native binary format

void PetscMatImpl::Save( const char *filename )
{
    PetscErrorCode ierr;
    PetscViewer viewer;

#if 1
    // output the matrix in binary format
    ierr = PetscViewerBinaryOpen( MPI_COMM_SELF, filename, PETSC_FILE_CREATE, &viewer );CHKERRQ(ierr);
    ierr = PetscViewerSetFormat( viewer, PETSC_VIEWER_BINARY_NATIVE );CHKERRQ(ierr);
#else
    // output the matrix in matlab format
    ierr = PetscViewerASCIIOpen( MPI_COMM_SELF, filename, &viewer );CHKERRQ(ierr);
    ierr = PetscViewerSetFormat( viewer, PETSC_VIEWER_ASCII_MATLAB);CHKERRQ(ierr);
#endif

    ierr = MatView(m_PetscMat,viewer);CHKERRQ(ierr);
    ierr = PetscViewerDestroy(viewer);CHKERRQ(ierr);
}

//////////////////////////////////////////////////////////////////////////////
// Load - Load a file in Petsc's native binary format

void PetscMatImpl::Load( const char *filename )
{
    Clear();

    PetscErrorCode ierr;
    PetscViewer viewer;
    ierr = PetscViewerBinaryOpen( MPI_COMM_SELF, filename, PETSC_FILE_RDONLY, &viewer );CHKERRQ(ierr);
    ierr = MatLoad( viewer, MATSEQAIJ, &m_PetscMat );CHKERRQ(ierr);
    ierr = PetscViewerDestroy(viewer);CHKERRQ(ierr);

    PetscTruth b;
    ierr = MatIsSymmetric( m_PetscMat, 1e-3, &b );CHKERRQ(ierr);
    if( b )
        m_bSymmetric = true;

    // Initialize the solver
    InitKSP();
}

//////////////////////////////////////////////////////////////////////////////
// Set - Sets the value of an entry. Creates the entry if it doesn't exist,
//       overwrites the entry if it does exist.

void PetscMatImpl::Set( int i, int j, real value )
{
    assert( i < m_numRows && j < m_numColumns );

    m_rows[i][j] = value;

    // if it's supposed to be symmetric, set the entry in the other half also
    if( m_bSymmetric && i != j )
        m_rows[j][i] = value;
}

//////////////////////////////////////////////////////////////////////////////
// Add - Adds the value of an entry. If it doesn't exist, assumes it was 0

void PetscMatImpl::Add( int i, int j, real value )
{
    assert( i < m_numRows && j < m_numColumns );

    m_rows[i][j] += value;

    // if it's supposed to be symmetric, set the entry in the other half also
    if( m_bSymmetric && i != j )
        m_rows[j][i] += value;
}

//////////////////////////////////////////////////////////////////////////////
// Multiply - A Matrix-Vector multiply.  product := this * vec

void PetscMatImpl::Multiply( const VectorImpl *vec, VectorImpl *product ) const
{
    PetscErrorCode ierr = MatMult( m_PetscMat,
        static_cast<const PetscVectorImpl*>(vec)->m_PetscVec,
        static_cast<PetscVectorImpl*>(product)->m_PetscVec );CHKERRQ(ierr);
}

//////////////////////////////////////////////////////////////////////////////
// Solve - A preconditioned Conjugate Gradient solver

bool PetscMatImpl::Solve( const VectorImpl *b, VectorImpl *x, int, real ) const
{
    //assert( m_bSymmetric && x.size() == m_numColumns && b.size() == m_numRows );

    PetscErrorCode ierr;

    ierr = KSPSolve( m_PetscKSP,
        static_cast<const PetscVectorImpl*>(b)->m_PetscVec,
        static_cast<PetscVectorImpl*>(x)->m_PetscVec );CHKERRQ(ierr);

    KSPConvergedReason reason;
    ierr = KSPGetConvergedReason(m_PetscKSP,&reason);CHKERRQ(ierr);
    if( reason < 0 )
    {
        LOG(warn) << "Solver failed to converge. Reason = " << reason;
        return false;
    }

    if( m_bLogConvergenceStats )
    {
        PetscInt its;
        KSPGetIterationNumber( m_PetscKSP, &its );
        LOG(info) << "Solver converged in " << its << " iterations. Reason: " << reason;
    }

    return true;
}


#endif  // COMPILE_WITH_PETSC
