//////////////////////////////////////////////////////////////////////////////
// LinearAlgebra.h
// Declares the Vector class and the Sparse Matrix class.
//////////////////////////////////////////////////////////////////////////////

#ifndef _LINEAR_ALGEBRA_H_
#define _LINEAR_ALGEBRA_H_


//////////////////////////////////////////////////////////////////////////////
// Includes

#include "util/real.h"


//////////////////////////////////////////////////////////////////////////////
// Enums

enum PRECOND_TYPE_ENUM
{
    PRECOND_TYPE_IDENTITY,
    PRECOND_TYPE_DIAGONAL,
    PRECOND_TYPE_INCOMPLETE_CHOLESKY
};


//////////////////////////////////////////////////////////////////////////////
// class Vector

class Vector
{
    friend class SparseMatrix;

public:
    Vector( size_t size );
    ~Vector();

    void Assign  ( size_t size, double value );
    void Set     ( double value );
    void Copy    ( const Vector &from );  // this Vector must have the same size as from
    void Swap    ( Vector &v );

    double & operator [] ( size_t i );
    const double & operator [] ( size_t i ) const;

    // These functions operate component-wise. The passed-in Vector is one of
    // the operands and the 'this' Vector serves as both an operand and the
    // destination.
    void Add       ( const Vector &v );
    void Subtract  ( const Vector &v );
    void Multiply  ( const Vector &v );
    void Multiply  ( real f );
    void Divide    ( const Vector &v );
    void Negate    ( void );
    void MAD       ( real f, const Vector &v );  // f*this + v
    void MAD       ( const Vector &v, real f );  // f*v + this

    real DotProduct( const Vector &v ) const;

private:
    // no copying
    Vector( const Vector & );
    Vector& operator=( const Vector & );

    class VectorImpl *m_impl;
};



//////////////////////////////////////////////////////////////////////////////
// class SparseMatrix

class SparseMatrix
{
public:
    SparseMatrix();
    SparseMatrix( int numRows, int numColumns );

    ~SparseMatrix();

    // Clears any current data
    void BeginConstruction( int numRows, int numColumns, bool bSymmetric = false );
    void Set( int i, int j, real value );
    void Add( int i, int j, real value );
    void EndConstruction( PRECOND_TYPE_ENUM type = PRECOND_TYPE_INCOMPLETE_CHOLESKY, real dropTolerance = (real)1.0e-1 );

    void Save( const char *filename );
    void Load( const char *filename );

    void Multiply( const Vector &vec, Vector &product ) const;

    bool Solve( const Vector &b, Vector &x, int maxIterations = 200, real epsilon = 1.0e-3 ) const;

    void LogConvergenceStats( bool enable );

private:
    // no copying
    SparseMatrix( const SparseMatrix & );
    SparseMatrix& operator=( const SparseMatrix & );

    class MatImpl *m_impl;
};


#endif  // _LINEAR_ALGEBRA_H_