
//////////////////////////////////////////////////////////////////////////////
// Includes

#include "simplicialComplex/meshFactory.h"
#include "simplicialComplex/vertex.h"
#include "simplicialComplex/edge.h"
#include "simplicialComplex/face.h"
#include "simplicialComplex/tet.h"
#include "simplicialComplex/iterator.h"
#include "simplicialComplex/vertexEdgeIterator.h"
#include "simplicialComplex/edgeFaceIterator.h"
#include "util/geometry.h"
#include "util/assert.h"

#include <algorithm>


//////////////////////////////////////////////////////////////////////////////
// Default Constructor

MeshFactory<3>::MeshFactory( SimplicialComplex<3>& mesh ) : sc(mesh)
{
    // empty
}

//////////////////////////////////////////////////////////////////////////////
// Destructor

MeshFactory<3>::~MeshFactory()
{
    // empty
}

//////////////////////////////////////////////////////////////////////////////
// Build - Given tets defined as 4-tuples, construct the edge and face sets,
//         and the boundary operators. We need the positions because we're
//         going to use the volume form to orient the tets.

bool MeshFactory<3>::Build( const std::vector< vec3 >& vertexPositions,
                            const std::vector< vec<4,int> >& tets )
{
    // Create Simplex objects out of the vertex positions
    sc.m_vertices.clear();
    sc.m_vertices.reserve( vertexPositions.size() );
    for( int i = 0; i < (int)vertexPositions.size(); ++i )
        if( !MakeVertex( i, vertexPositions[i] ) )
            return false;

    // Create Simplex objects out of the tet tuples
    sc.m_tets.clear();
    sc.m_tets.reserve( tets.size() );
    for( std::vector< vec<4,int> >::const_iterator tet = tets.begin(); tet != tets.end(); ++tet )
        if( !MakeTet( (*tet)[0], (*tet)[1], (*tet)[2], (*tet)[3] ) )
            return false;

    // Assign tet indices
    for( size_t i = 0; i < sc.NumTets(); ++i )
        sc.m_tets[i].index = i;

    // Fill in m_faces and m_edges
    if( !BuildFaces() || !BuildEdges() )
        return false;

    // Fill in the adjacency matrices
    if( !BuildAdjacencyVE() || !BuildAdjacencyEF() || !BuildAdjacencyFT() || !BuildAdjacencyET() )
        return false;


    // VERIFY!


    // Calculate all the primal and dual volumes
    if( !CalculateVolumes() )
        return false;


    LOG(dbg) << "vertices: " << (int)sc.NumVerts();
    LOG(dbg) << "edges:    " << (int)sc.NumEdges();
    LOG(dbg) << "faces:    " << (int)sc.NumFaces();
    LOG(dbg) << "tets:     " << (int)sc.NumTets();

    return true;
}

//////////////////////////////////////////////////////////////////////////////
// MakeVertex - Creates a 0-simplex.

bool MeshFactory<3>::MakeVertex( int v0, const vec3& position )
{
    Simplex<0> simplex = MakeSimplex( v0 );
    simplex.index = v0;
    simplex.center = position;
    simplex.volume = (real)1.0;
    simplex.dualVolume = (real)1.0;

    sc.m_vertices.push_back( simplex );

    return true;
}

//////////////////////////////////////////////////////////////////////////////
// MakeTet - Given a set of 4 integers, make a 3-simplex out of it, orient it,
//           and add it to the tet list.

bool MeshFactory<3>::MakeTet( int v0, int v1, int v2, int v3 )
{
    int numVerts = (int)sc.NumVerts();

    if( v0 < 0 || v0 >= numVerts || v1 < 0 || v1 >= numVerts ||
        v2 < 0 || v2 >= numVerts || v3 < 0 || v3 >= numVerts )
    {
        LOG(err) << "Invalid sc input. Tet's vertex index is out of range.";
        LOG(err) << "    Tet is [" << v0 << ", " << v1 << ", " << v2 << ", " << v3 << "]";
        LOG(err) << "    Maximum index is " << numVerts - 1;
        return false;
    }

    Simplex<3> simplex = MakeSimplex( v0, v1, v2, v3 );

    vec3 verts[4] = { sc.m_vertices[v0].center,
                      sc.m_vertices[v1].center,
                      sc.m_vertices[v2].center,
                      sc.m_vertices[v3].center };

    simplex.center = Circumcenter( verts[0], verts[1], verts[2], verts[3] );
    simplex.volume = Volume( verts[0], verts[1], verts[2], verts[3] );
    simplex.dualVolume = (real)1.0;

    // Orient the tet according to its volume form
    if( simplex.volume < 0 )
    {
        std::swap( simplex.v[2], simplex.v[3] );
        simplex.volume = -simplex.volume;
    }

    sc.m_tets.push_back( simplex );

    return true;
}

//////////////////////////////////////////////////////////////////////////////
// BuildFaces - Populates m_faces. Assumes that m_vertices and m_tets are
//              already filled in.

bool MeshFactory<3>::BuildFaces()
{
    // Iterate over the tets and add the four faces of each.
    // There will be duplicates, but they will be removed at the end.
    for( S3ListIter tet = sc.m_tets.begin(); tet != sc.m_tets.end(); ++tet )
    {
        Simplex<2> newFaces[4] = { MakeSimplex( tet->Get(1), tet->Get(2), tet->Get(3) ),
                                   MakeSimplex( tet->Get(0), tet->Get(2), tet->Get(3) ),
                                   MakeSimplex( tet->Get(0), tet->Get(1), tet->Get(3) ),
                                   MakeSimplex( tet->Get(0), tet->Get(1), tet->Get(2) ) };

        // We (arbitrarily) define the intrinsic orientation of faces to
        // be when the indices are sorted.
        newFaces[0].ResetSign();
        newFaces[1].ResetSign();
        newFaces[2].ResetSign();
        newFaces[3].ResetSign();

        sc.m_faces.insert( sc.m_faces.end(), newFaces, newFaces + 4 );
    }

    // Remove duplicates
    sort( sc.m_faces.begin(), sc.m_faces.end() );
    sc.m_faces.erase( unique( sc.m_faces.begin(), sc.m_faces.end() ), sc.m_faces.end() );

    // Trip excess capacity
    SimplicialComplex<3>::S2List(sc.m_faces).swap(sc.m_faces);

    // Assign indices
    for( size_t i = 0; i < sc.NumFaces(); ++i )
        sc.m_faces[i].index = i;

    return true;
}

//////////////////////////////////////////////////////////////////////////////
// BuildEdges - Populates m_edges. Assumes that m_faces is already filled in.

bool MeshFactory<3>::BuildEdges()
{
    // Iterate over the faces and add the three edges of each.
    // There will be duplicates, but they will be removed at the end.
    for( S2ListIter face = sc.m_faces.begin(); face != sc.m_faces.end(); ++face )
    {
        Simplex<1> newEdges[3] = { MakeSimplex( face->Get(1), face->Get(2) ),
                                   MakeSimplex( face->Get(2), face->Get(0) ),
                                   MakeSimplex( face->Get(0), face->Get(1) ) };

        // We (arbitrarily) define the intrinsic orientation of edges to
        // be when the indices are sorted.
        newEdges[0].ResetSign();
        newEdges[1].ResetSign();
        newEdges[2].ResetSign();

        sc.m_edges.insert( sc.m_edges.end(), newEdges, newEdges + 3 );
    }

    // Remove duplicates
    sort( sc.m_edges.begin(), sc.m_edges.end() );
    sc.m_edges.erase( unique( sc.m_edges.begin(), sc.m_edges.end() ), sc.m_edges.end() );

    // Trip excess capacity
    SimplicialComplex<3>::S1List(sc.m_edges).swap(sc.m_edges);

    // Assign indices
    for( size_t i = 0; i < sc.NumEdges(); ++i )
        sc.m_edges[i].index = i;

    return true;
}

//////////////////////////////////////////////////////////////////////////////
// BuildAdjacencyVE

bool MeshFactory<3>::BuildAdjacencyVE()
{
    sc.m_EVmatrix.resize( sc.NumEdges() );
    sc.m_VEmatrix.resize( sc.NumVerts() );

    // Iterate over the edges and add the two vertices of each.
    for( S1ListIter edge = sc.m_edges.begin(); edge != sc.m_edges.end(); ++edge )
    {
        sc.m_EVmatrix[ edge->index ][ 0 ] = edge->Get(0);
        sc.m_EVmatrix[ edge->index ][ 1 ] = edge->Get(1);

        sc.m_VEmatrix[ edge->Get(0) ].push_back( SignedIndex( edge->index, -1 ) );
        sc.m_VEmatrix[ edge->Get(1) ].push_back( SignedIndex( edge->index,  1 ) );
    }

    return true;
}

//////////////////////////////////////////////////////////////////////////////
// BuildAdjacencyEF

bool MeshFactory<3>::BuildAdjacencyEF()
{
    sc.m_FEmatrix.resize( sc.NumFaces() );
    sc.m_EFmatrix.resize( sc.NumEdges() );

    // Iterate over the faces and add the three edges of each.
    for( S2ListIter face = sc.m_faces.begin(); face != sc.m_faces.end(); ++face )
    {
        Simplex<1> edges[3] = { MakeSimplex( face->Get(1), face->Get(2) ),
                                MakeSimplex( face->Get(2), face->Get(0) ),
                                MakeSimplex( face->Get(0), face->Get(1) ) };

        // find the index of the face with a binary search through the sorted face list
        S1ListIter edgeIter[3] = {
            std::lower_bound( sc.m_edges.begin(), sc.m_edges.end(), edges[0] ),
            std::lower_bound( sc.m_edges.begin(), sc.m_edges.end(), edges[1] ),
            std::lower_bound( sc.m_edges.begin(), sc.m_edges.end(), edges[2] ) };

        int sign[3] = { ( IsSameSign(edges[0], *edgeIter[0]) ) ? 1 : -1,
                        ( IsSameSign(edges[1], *edgeIter[1]) ) ? 1 : -1,
                        ( IsSameSign(edges[2], *edgeIter[2]) ) ? 1 : -1 };

        sc.m_FEmatrix[ face->index ][ 0 ] = SignedIndex( edgeIter[0]->index, sign[0] );
        sc.m_FEmatrix[ face->index ][ 1 ] = SignedIndex( edgeIter[1]->index, sign[1] );
        sc.m_FEmatrix[ face->index ][ 2 ] = SignedIndex( edgeIter[2]->index, sign[2] );

        sc.m_EFmatrix[ edgeIter[0]->index ].push_back( SignedIndex( face->index, sign[0] ) );
        sc.m_EFmatrix[ edgeIter[1]->index ].push_back( SignedIndex( face->index, sign[1] ) );
        sc.m_EFmatrix[ edgeIter[2]->index ].push_back( SignedIndex( face->index, sign[2] ) );
    }

    return true;
}

//////////////////////////////////////////////////////////////////////////////
// BuildAdjacencyFT

bool MeshFactory<3>::BuildAdjacencyFT()
{
    sc.m_TFmatrix.resize( sc.NumTets() );
    sc.m_FTmatrix.resize( sc.NumFaces() );

    // Initialize the FT matrix to all -1s
    for( S2ListIter face = sc.m_faces.begin(); face != sc.m_faces.end(); ++face )
        sc.m_FTmatrix[face->index][0] = sc.m_FTmatrix[face->index][1] = size_t(-1);

    // Iterate over the tets and add the four faces of each.
    for( S3ListIter tet = sc.m_tets.begin(); tet != sc.m_tets.end(); ++tet )
    {
        Simplex<2> faces[4] = {  MakeSimplex( tet->Get(1), tet->Get(2), tet->Get(3) ),
                                -MakeSimplex( tet->Get(0), tet->Get(2), tet->Get(3) ),
                                 MakeSimplex( tet->Get(0), tet->Get(1), tet->Get(3) ),
                                -MakeSimplex( tet->Get(0), tet->Get(1), tet->Get(2) ) };

        // find the index of the face with a binary search through the sorted face list
        S2ListIter faceIter[4] = {
            std::lower_bound( sc.m_faces.begin(), sc.m_faces.end(), faces[0] ),
            std::lower_bound( sc.m_faces.begin(), sc.m_faces.end(), faces[1] ),
            std::lower_bound( sc.m_faces.begin(), sc.m_faces.end(), faces[2] ),
            std::lower_bound( sc.m_faces.begin(), sc.m_faces.end(), faces[3] ) };

        int sign[4] = { ( IsSameSign(faces[0], *faceIter[0]) ) ? 1 : -1,
                        ( IsSameSign(faces[1], *faceIter[1]) ) ? 1 : -1,
                        ( IsSameSign(faces[2], *faceIter[2]) ) ? 1 : -1,
                        ( IsSameSign(faces[3], *faceIter[3]) ) ? 1 : -1 };

        sc.m_TFmatrix[ tet->index ][ 0 ] = SignedIndex( faceIter[0]->index, sign[0] );
        sc.m_TFmatrix[ tet->index ][ 1 ] = SignedIndex( faceIter[1]->index, sign[1] );
        sc.m_TFmatrix[ tet->index ][ 2 ] = SignedIndex( faceIter[2]->index, sign[2] );
        sc.m_TFmatrix[ tet->index ][ 3 ] = SignedIndex( faceIter[3]->index, sign[3] );

        // the negative one is first, the positive one is second
        sc.m_FTmatrix[ faceIter[0]->index ][ (sign[0] == 1) ? 1 : 0 ] = tet->index;
        sc.m_FTmatrix[ faceIter[1]->index ][ (sign[1] == 1) ? 1 : 0 ] = tet->index;
        sc.m_FTmatrix[ faceIter[2]->index ][ (sign[2] == 1) ? 1 : 0 ] = tet->index;
        sc.m_FTmatrix[ faceIter[3]->index ][ (sign[3] == 1) ? 1 : 0 ] = tet->index;
    }

    return true;
}

//////////////////////////////////////////////////////////////////////////////
// BuildAdjacencyET

bool MeshFactory<3>::BuildAdjacencyET()
{
    sc.m_ETmatrix.resize( sc.NumEdges() );

    Tet t0, t1;
    for( SimplexIterator<Edge> edge = sc.Edges(); edge; ++edge )
    {
        std::vector<size_t> & curMatrixRow = sc.m_ETmatrix[ edge->Index() ];

        for( EdgeFaceIterator face = edge->Faces(); face; ++face )
        {
            face->Tets( t0, t1 );
            curMatrixRow.push_back( t0.Index() );
            curMatrixRow.push_back( t1.Index() );
        }

        // remove duplicates
        sort( curMatrixRow.begin(), curMatrixRow.end() );
        curMatrixRow.erase( unique( curMatrixRow.begin(), curMatrixRow.end() ), curMatrixRow.end() );
    }

    return true;
}

//////////////////////////////////////////////////////////////////////////////
// CalculateVolumes - Primal and dual. Assumes all the adjacency matrices are
//                    filled in.

bool MeshFactory<3>::CalculateVolumes()
{
    // It must be done in this order (faces, then edges, then vertices) because
    // each uses the results of the previous calculations

    // Faces
    for( SimplexIterator<Face> face = sc.Faces(); face; ++face )
    {
        Simplex<2>& faceSimplex = sc.m_faces[ face->Index() ];

        vec3 v[3];
        face->Vertices( v[0], v[1], v[2] );
        faceSimplex.center = Circumcenter( v[0], v[1], v[2] );
        faceSimplex.volume = Area( v[0], v[1], v[2] );

        Tet t[2];
        face->Tets( t[0], t[1] );
        faceSimplex.dualVolume = Length( t[0] ? t[0].Center() : face->Center(),
                                         t[1] ? t[1].Center() : face->Center() );
    }

    // Edges
    for( SimplexIterator<Edge> edge = sc.Edges(); edge; ++edge )
    {
        Simplex<1>& edgeSimplex = sc.m_edges[ edge->Index() ];

        vec3 v[2];
        edge->Vertices( v[0], v[1] );
        edgeSimplex.center = (real)0.5 * (v[0] + v[1]);
        edgeSimplex.volume = Length( v[0], v[1] );

        edgeSimplex.dualVolume = (real)0.0;
        for( EdgeFaceIterator face = edge->Faces(); face; ++face )
        {
            real toFaceCenter = Length( edgeSimplex.center, face->Center() );
            edgeSimplex.dualVolume += (real)0.5 * toFaceCenter * face->DualLength();
        }
    }

    // Vertices
    for( SimplexIterator<Vertex> vertex = sc.Vertices(); vertex; ++vertex )
    {
        Simplex<0>& vertexSimplex = sc.m_vertices[ vertex->Index() ];
        vertexSimplex.dualVolume = (real)0.0;
        for( VertexEdgeIterator edge = vertex->Edges(); edge; ++edge )
            vertexSimplex.dualVolume += ((real)1.0/(real)6.0) * edge->Length() * edge->DualArea();
    }

    return true;
}

