#ifndef _SIMPLEX_H_
#define _SIMPLEX_H_


//////////////////////////////////////////////////////////////////////////////
// Includes

#include "util/linear.h"


//////////////////////////////////////////////////////////////////////////////
// class Simplex
// A template class for N-simplices. A simplex is implemented as an N-tuple
// of vertex indices, with a sign denoting the orientation of the simplex
// relative to the order of the indices in the tuple.

template<int N>
class Simplex
{
    friend Simplex<0> MakeSimplex( int );
    friend Simplex<1> MakeSimplex( int, int );
    friend Simplex<2> MakeSimplex( int, int, int );
    friend Simplex<3> MakeSimplex( int, int, int, int );

    template<int N> friend class MeshFactory;

public:
    // The tuples are always stored sorted, with the sign taking
    // into account the permutation. Or you can optionally set
    // sort to false, then the order stays the way it came in and
    // the caller is responsible for orienting it.
    Simplex( int _v[N], bool sort = true ) : index(size_t(-1)), center(0,0,0), volume(0), dualVolume(0)
    {
        sign = 1;
        for( int i = 0; i <= N; ++i )
            v[i] = _v[i];

        if( sort )
        {
            // bubble sort
            int tmp;
            for( int i = 0; i < N; ++i )
            {
                for( int j = 0; j < N-i; ++j )
                {
                    if( v[j+1] < v[j] )
                    {
                        tmp = v[j];
                        v[j] = v[j+1];
                        v[j+1] = tmp;
                        sign *= -1;
                    }
                }
            }
        }
    }

    Simplex( const Simplex& rhs )
    {
        *this = rhs;
    }

    Simplex& operator=( const Simplex& rhs )
    {
        sign = rhs.sign;
        for( int i = 0; i <= N; ++i )
            v[i] = rhs.v[i];
        index = rhs.index;
        center = rhs.center;
        volume = rhs.volume;
        dualVolume = rhs.dualVolume;
        return (*this);
    }

    int  Get       ( int i ) const  { return v[i]; }
    bool IsNegative( void )  const  { return (sign < 0); }
    void Negate    ( void )         { sign = -sign; }
    void ResetSign ( void )         { sign = 1; }

    size_t Index     ( void )  const  { return index; }
    vec3   Center    ( void )  const  { return center; }
    real   Volume    ( void )  const  { return volume; }
    real   DualVolume( void )  const  { return dualVolume; }

private:
    int sign;
    int v[N+1];

    size_t index;
    vec3   center;
    real   volume;
    real   dualVolume;
};


// None of the comparison operators take into account the sign,
// they just test whether they refer to the same (absolute) simplex.
// If you want to know whether they are also oriented the same,
// use IsSameSign()

template<int N>
inline bool operator==( const Simplex<N>& lhs, const Simplex<N>& rhs )
{
    for( int i = 0; i <= N; ++i )
        if( lhs.Get(i) != rhs.Get(i) )
            return false;
    return true;
}

template<int N>
inline bool operator!=( const Simplex<N>& lhs, const Simplex<N>& rhs )
{
    return (!(lhs == rhs));
}

template<int N>
inline bool operator<( const Simplex<N>& lhs, const Simplex<N>& rhs )
{
    for( int i = 0; i < N; ++i )
    {
        if( lhs.Get(i) < rhs.Get(i) )
            return true;
        else if( lhs.Get(i) > rhs.Get(i) )
            return false;
    }

    return (lhs.Get(N) < rhs.Get(N));
}

template<int N>
inline bool operator>( const Simplex<N>& lhs, const Simplex<N>& rhs )
{
    return (rhs < lhs);
}

template<int N>
inline bool operator<=( const Simplex<N>& lhs, const Simplex<N>& rhs )
{
    return (!(rhs < lhs));
}

template<int N>
inline bool operator>=( const Simplex<N>& lhs, const Simplex<N>& rhs )
{
    return (!(lhs < rhs));
}

template<int N>
inline Simplex<N> operator-( const Simplex<N>& simplex )
{
    Simplex<N> negated( simplex );
    negated.Negate();
    return negated;
}

template<int N>
inline bool IsSameSign( const Simplex<N>& lhs, const Simplex<N>& rhs )
{
    return (lhs.IsNegative() == rhs.IsNegative());
}


inline Simplex<0> MakeSimplex( int v0 )
{
    int v[1] = {v0};
    return Simplex<0>( v );
}

inline Simplex<1> MakeSimplex( int v0, int v1 )
{
    int v[2] = {v0, v1};
    return Simplex<1>( v );
}

inline Simplex<2> MakeSimplex( int v0, int v1, int v2 )
{
    int v[3] = {v0, v1, v2};
    return Simplex<2>( v );
}

inline Simplex<3> MakeSimplex( int v0, int v1, int v2, int v3 )
{
    int v[4] = {v0, v1, v2, v3};
    return Simplex<3>( v, false );
}


#endif  // _SIMPLEX_H_
