#include "util/log/log_impl.h"
#include <map>
#include <vector>
#include <algorithm>

namespace ns_log { 

struct appender_info
{
    appender_info(const std::string & spec, appender_func func) : spec(spec), func(func) {}
    std::string spec;
    appender_func func;
};

struct modifier_info
{
    modifier_info(const std::string & spec, modifier_func func) : spec(spec), func(func) {}
    std::string spec;
    modifier_func func;
};

struct enabled_info
{
    enabled_info(const std::string & spec, bool is) : spec(spec), is_enabled(is) {}
    std::string spec;
    bool is_enabled;
};

typedef std::map<std::string, logger_impl*> logger_map;

struct log_manager
{
    log_manager() {}
    ~log_manager() {}

    logger_map m_logs;
    std::vector<appender_info> m_appenders;
    std::vector<modifier_info> m_modifiers;
    std::vector<enabled_info> m_enabled;
};

log_manager & manager()
{
    static log_manager m;
    return m;
}

// returns true if the string matches the specification
// empty string and '*' match everything
bool name_matches_spec(const std::string & log_name, const std::string & spec)
{
    if ( (spec == "*") || spec.empty() ) 
        return true; // matches all

    std::vector<std::string> searches;
    std::string::size_type pos = 0, next;
    while( pos < spec.size() )
    {
        next = spec.find('*', pos);
        if( next == std::string::npos ) next = spec.size();
        if( next - pos > 0 )
            searches.push_back( spec.substr(pos, next - pos) );
        pos = next + 1;
    }
    if( searches.empty() )
        return true; // string was only wildcards

    std::string::size_type at = 0;
    for( std::vector<std::string>::const_iterator iter = searches.begin(); iter != searches.end(); ++iter )
    {
        std::string::size_type next = log_name.find(*iter, at);
        if( next == std::string::npos )
            return false; // doesn't match
        at = next + iter->length();
    }

    // if fixed, when search starts/ends, the log_name must start/end
    if( *spec.begin() != '*' && log_name.find(*searches.begin()) != 0 )
        return false; // log_name should be like "word*" (start with 'word')
    if( *spec.rbegin() != '*' && at != log_name.length() )
        return false; // log_name should be like "*word" (end in 'word')

    return true;
}


void add_appender(const std::string & logs_spec, appender_func f)
{
    // every operation reflects to all logs that match logs_spec (even logs that will be created in the future)
    appender_info info( logs_spec, f );
    manager().m_appenders.push_back( info );

    for( logger_map::iterator iter = manager().m_logs.begin(); iter != manager().m_logs.end(); ++iter )
        if( name_matches_spec( iter->second->name, logs_spec ) )
            iter->second->add_appender(f);
}

void add_modifier(const std::string & logs_spec, modifier_func f)
{
    // every operation reflects to all logs that match logs_spec (even logs that will be created in the future)
    modifier_info info( logs_spec, f );
    manager().m_modifiers.push_back( info );

    for( logger_map::iterator iter = manager().m_logs.begin(); iter != manager().m_logs.end(); ++iter )
        if( name_matches_spec( iter->second->name, logs_spec ) )
            iter->second->add_modifier(f);
}

// refreshes each log's "enabled" property. Called after an "enable_logs" or "disable_logs" has happened
void refresh_log_enabled()
{
    for( logger_map::iterator iter = manager().m_logs.begin(); iter != manager().m_logs.end(); ++iter )
    {
        bool is_enabled = true;
        for( std::vector<enabled_info>::iterator b = manager().m_enabled.begin(); b != manager().m_enabled.end(); ++b )
            if( name_matches_spec( iter->second->name, b->spec ) )
                is_enabled = b->is_enabled;
        iter->second->enable(is_enabled);
    }
}

void enable_logs(const std::string & logs_spec)
{
    // every operation reflects to all logs that match logs_spec (even logs that will be created in the future)
    manager().m_enabled.push_back( enabled_info(logs_spec, true) );
    refresh_log_enabled();
}

void disable_logs(const std::string & logs_spec)
{
    // every operation reflects to all logs that match logs_spec (even logs that will be created in the future)
    manager().m_enabled.push_back( enabled_info(logs_spec, false) );
    refresh_log_enabled();
}

logger_impl* find_log_by_name(const std::string & log_name)
{
    logger_map::iterator found = manager().m_logs.find(log_name);
    if( found != manager().m_logs.end() )
        return found->second;

    // this log does not yet exist, let's create it now
    // the pointer is not deleted, it's kept forever in manager().m_logs
    logger_impl *new_log = new logger_impl(log_name);

    // add appenders
    for( std::vector<appender_info>::iterator iter = manager().m_appenders.begin(); iter != manager().m_appenders.end(); ++iter )
        if( name_matches_spec( log_name, iter->spec ) )
            new_log->add_appender(iter->func);

    // add modifiers
    for( std::vector<modifier_info>::iterator iter = manager().m_modifiers.begin(); iter != manager().m_modifiers.end(); ++iter )
        if( name_matches_spec( log_name, iter->spec ) )
            new_log->add_modifier(iter->func);

    // find out if it's enabled or disabled
    bool is_enabled = false;
    for( std::vector<enabled_info>::iterator iter = manager().m_enabled.begin(); iter != manager().m_enabled.end(); ++iter )
        if( name_matches_spec( log_name, iter->spec ) )
            is_enabled = iter->is_enabled;
    new_log->enable( is_enabled );

    manager().m_logs.insert( std::make_pair(log_name, new_log) );
    return new_log;
}


}  // namespace ns_log
