#ifndef _REAL_H_
#define _REAL_H_

#undef max
#undef min

#include <cstdlib>
#include <limits>


// comment this out for double precision
#define REAL_IS_FLOAT

#ifdef REAL_IS_FLOAT
    typedef float real;
#else
    typedef double real;
#endif


extern const real E;
extern const real EPSILON;
extern const real PI;
extern const real SQRT_2;
extern const real SQRT_3;


/** @addtogroup util */
// @{

inline real max_real() {
    return std::numeric_limits<real>::max();
}

inline real infinity() {
    return std::numeric_limits<real>::infinity();
}

inline real NaN() {
    return std::numeric_limits<real>::quiet_NaN();
}

inline real random() {
    return (real(rand()) / real(RAND_MAX));
}

inline void randomize(unsigned int seed) {
    srand(seed);
}

inline bool gt(const real& x, const real& y) {
    return (x - y > EPSILON);
}

inline bool geq(const real& x, const real& y) {
    return (x - y > -EPSILON);
}

inline bool eq(const real& x, const real& y) {
    return (x - y < EPSILON && x - y > -EPSILON);
}

inline bool neq(const real& x, const real& y) {
    return (x - y > EPSILON || x - y < -EPSILON);
}

inline bool leq(const real& x, const real& y) {
    return (x - y < EPSILON);
}

inline bool lt(const real& x, const real& y) {
    return (x - y < -EPSILON);
}

inline real clamp(const real& x, const real& min, const real& max) {
    return (x > min ? (x < max ? x : max) : min);
}

inline bool isInfinite(const real& x) {
    return (fabs(x) == infinity());
}

inline bool isNaN(const real& x) {
    return (x != x);
}

inline real max(const real& x, const real& y) {
    return (x >= y ? x : y);
}

inline real max(const real& x, const real& y, const real& z) {
    return (max(max(x, y), z));
}

inline real min(const real& x, const real& y) {
    return (x <= y ? x : y);
}

inline real min(const real& x, const real& y, const real& z) {
    return (min(min(x, y), z));
}

inline real round(const real& x) {
    return static_cast<real>(floor(x + 0.5));
}

inline real sign(const real& x) {
    return (x == 0 ? real(0) : (x > 0 ? real(1) : real(-1)));
}

inline real sqr(const real& x) {
    return (x * x);
}

inline real toDegrees(const real& x) {
    return (x * (real)57.2957795130823208767981548141052);
}

inline real toRadians(const real& x) {
    return (x * (real)0.0174532925199432957692369076848861);
}

// @}

#endif  // _REAL_H_
