#include "util/tokenizer.h"
#include "util/assert.h"
#include <cstdio>
#include <sstream>

Tokenizer::Tokenizer(std::istream& in) : _in(&in), _line(1), _owner(false), _putback(false), _quote(0) {
    *_token = 0;
    resetSyntax();
}

Tokenizer::Tokenizer(const std::string& string) : _line(1), _putback(false) {
    _in = new std::stringstream(string);
    _owner = true;
    *_token = 0;
    resetSyntax();
}

Tokenizer::~Tokenizer() {
    if (_owner)
        delete (_in);
}

Tokenizer& Tokenizer::next() {
    if (_putback) {
        _putback = false;
    } else if (!*_in) {
        *_token = 0;
    } else {
        char c;
        char* w = _token;

        do {
            _in->get(c);
            if (!*_in) {
                *w = 0;
                return (*this);
            }

            if (c == '\n')
                _line++;
        } while (_isWhitespace(c));

        if (_isQuote(c)) {
            _quote = c;
        } else {
            _quote = 0;
            *w++ = c;
        }

        if (!_isOrdinary(c)) {
            while (_in->get(c)) {
                if (!_quote && !_isWord(c)) {
                    _in->putback(c);
                    break;
                } else if (c == _quote) {
                    break;
                } else {
                    if (c == '\n') {
                        _line++;
                        c = ' ';
                    }

                    *w++ = c;
                    if (w - _token == TOKEN_LENGTH)
                        break;
                }
            }
        }

        *w = 0;
    }

    return (*this);
}

Tokenizer& Tokenizer::nextLine() {
    char c;
    char* w = _token;
    bool skip = true;

    if (_putback) {
        w = &_token[strlen(_token)];
        skip = false;
        _putback = false;
    }

    do {
        while (w - _token < TOKEN_LENGTH) {
            _in->get(c);
            if (!*_in)
                break;

            if (c == '\n') {
                _line++;
                break;
            } else if (skip && _isWhitespace(c)) {
                continue;
            } else {
                *w++ = c;
                skip = false;
            }
        }

        smart_assert( w - _token != TOKEN_LENGTH )((int)(w - _token))(TOKEN_LENGTH).msg( "Line too long; truncated." );

        while (w != _token && _isWhitespace(*(w - 1)))
            w--;

        *w = 0;
    } while (w <= _token && *_in);

    return (*this);
}

void Tokenizer::ordinaryChar(unsigned char ordinary) {
    _type[ordinary] = ORDINARY;
}

void Tokenizer::ordinaryChars(const char* ordinary) {
    while (*ordinary)
        _type[*ordinary++] = ORDINARY;
}

void Tokenizer::ordinaryChars(unsigned char first, unsigned char last) {
    for (int c = first; c <= last; c++)
        _type[c] = ORDINARY;
}

void Tokenizer::quoteChar(unsigned char quote) {
    _type[quote] = QUOTE;
}

void Tokenizer::resetSyntax() {
    whitespaceChars(0, ' ');
    wordChars('!', 255);
    quoteChar('"');
}

void Tokenizer::whitespaceChar(unsigned char whitespace) {
    _type[whitespace] = WHITESPACE;
}

void Tokenizer::whitespaceChars(const char* whitespace) {
    while (*whitespace)
        _type[*whitespace++] = WHITESPACE;
}

void Tokenizer::whitespaceChars(unsigned char first, unsigned char last) {
    for (int c = first; c <= last; c++)
        _type[c] = WHITESPACE;
}

void Tokenizer::wordChar(unsigned char word) {
    _type[word] = WORD;
}

void Tokenizer::wordChars(const char* word) {
    while (*word)
        _type[*word++] = WORD;
}

void Tokenizer::wordChars(unsigned char first, unsigned char last) {
    for (int c = first; c <= last; c++)
        _type[c] = WORD;
}

std::ostream& operator<<(std::ostream& out, const Tokenizer& token) {
    if (token)
        return (out << token._token);
    else
        return (out << "EOF");
}
